import { __classPrivateFieldGet, __classPrivateFieldSet } from 'tslib';
import * as i0 from '@angular/core';
import { inject, ChangeDetectorRef, Directive, Injector, ErrorHandler } from '@angular/core';
import { isTruthy, wrapMethod } from '@s-libs/js-core';
import { mixInSubscriptionManager, delayOnMicrotaskQueue } from '@s-libs/rxjs-core';
import { Subject, BehaviorSubject, merge, of, tap, retry } from 'rxjs';
import { filter, distinctUntilChanged, map } from 'rxjs/operators';
import { NG_VALUE_ACCESSOR, NgControl, FormControl } from '@angular/forms';
import { noop, isEqual, omit, keys, size, flow, bindKey } from '@s-libs/micro-dash';

/**
 * Mixes in {@link InjectableSuperclass} as an additional superclass.
 *
 * ```ts
 * class MySubclass extends mixInInjectableSuperclass(MyOtherSuperclass) {
 *   subscribeWithAutoUnsubscribe(observable: Observable<any>) {
 *     this.subscribeTo(observable);
 *   }
 * }
 * ```
 */
// eslint-disable-next-line @typescript-eslint/explicit-function-return-type -- I'm not even sure how to describe the return type
function mixInInjectableSuperclass(Base) {
    var _destructionSubject, _a;
    return _a = class extends mixInSubscriptionManager(Base) {
            constructor() {
                super(...arguments);
                _destructionSubject.set(this, new Subject());
                /**
                 * An observable that emits once when this object is destroyed, then completes.
                 */
                this.destruction$ = __classPrivateFieldGet(this, _destructionSubject, "f").asObservable();
            }
            ngOnDestroy() {
                this.unsubscribe();
                __classPrivateFieldGet(this, _destructionSubject, "f").next();
                __classPrivateFieldGet(this, _destructionSubject, "f").complete();
            }
        },
        _destructionSubject = new WeakMap(),
        _a;
}
/**
 * Use as the superclass for anything managed by angular's dependency injection for care-free use of `subscribeTo()`. It simply calls `unsubscribe()` during `ngOnDestroy()`. If you override `ngOnDestroy()` in your subclass, be sure to invoke the super implementation.
 *
 * ```ts
 * @Injectable() // or @Component(), @Directive() or @Pipe(), but consider DirectiveSuperclass
 * class MyThing extends InjectableSuperclass {
 *   constructor(somethingObservable: Observable) {
 *     super();
 *     this.subscribeTo(somethingObservable);
 *   }
 *
 *   ngOnDestroy() {
 *     // if you override ngOnDestroy, be sure to call this too
 *     super.ngOnDestroy();
 *   }
 * }
 * ```
 */
class InjectableSuperclass extends mixInInjectableSuperclass(Object) {
}

var _DirectiveSuperclass_changeDetectorRef, _DirectiveSuperclass_onChangesRan$;
/**
 * Extend this when creating a directive (including a component, which is a kind of directive) to gain access to the helpers demonstrated below.
 *
 * ```ts
 * @Component({
 *   selector: "s-color-text",
 *   template: `
 *     <span [style.background]="color">{{ color }}</span>
 *   `,
 *   // note that `bindToInstance()` works even with OnPush change detection
 *   changeDetection: ChangeDetectionStrategy.OnPush,
 * })
 *  class ColorTextComponent extends DirectiveSuperclass {
 *   @Input() prefix?: string;
 *   @Input() prefix2?: string;
 *   color!: string;
 *
 *   constructor(@Inject("color$") color$: Observable<string>) {
 *     super();
 *
 *     // combine everything to calculate `color` and keep it up to date
 *     this.bindToInstance(
 *       "color",
 *       combineLatest(
 *         this.getInput$("prefix"),
 *         this.getInput$("prefix2"),
 *         color$,
 *       ).pipe(map((parts) => parts.filter((p) => p).join(""))),
 *     );
 *   }
 * }
 * ```
 */
// maybe this won't need the fake selector after https://github.com/angular/angular/issues/36427
// eslint-disable-next-line @angular-eslint/directive-class-suffix
class DirectiveSuperclass extends InjectableSuperclass {
    constructor() {
        super(...arguments);
        /**
         *  Emits the set of `@Input()` property names that change during each call to `ngOnChanges()`.
         */
        this.inputChanges$ = new Subject();
        _DirectiveSuperclass_changeDetectorRef.set(this, inject(ChangeDetectorRef));
        _DirectiveSuperclass_onChangesRan$.set(this, new BehaviorSubject(false));
    }
    ngOnChanges(changes) {
        this.inputChanges$.next(new Set(Object.getOwnPropertyNames(changes)));
        __classPrivateFieldGet(this, _DirectiveSuperclass_onChangesRan$, "f").next(true);
    }
    /**
     * @return an observable of the values for one of this directive's `@Input()` properties
     */
    getInput$(key) {
        // Should emit:
        //   - immediately if ngOnChanges was already called
        //   - on the first call to ngOnChanges
        //   - after a delay if ngOnChanges is never called (when nothing is bound to the directive)
        //   - when the value actually changes
        return merge(__classPrivateFieldGet(this, _DirectiveSuperclass_onChangesRan$, "f").pipe(filter(isTruthy), distinctUntilChanged()), this.inputChanges$, of(0).pipe(delayOnMicrotaskQueue())).pipe(map(() => this[key]), distinctUntilChanged());
    }
    /**
     * Binds an observable to one of this directive's instance variables. When the observable emits the instance variable will be updated, and change detection will be triggered to propagate any changes. Use this an an alternative to repeating `| async` multiple times in your template.
     */
    bindToInstance(key, value$) {
        this.subscribeTo(value$, (value) => {
            this[key] = value;
            __classPrivateFieldGet(this, _DirectiveSuperclass_changeDetectorRef, "f").markForCheck();
        });
    }
}
_DirectiveSuperclass_changeDetectorRef = new WeakMap(), _DirectiveSuperclass_onChangesRan$ = new WeakMap();
DirectiveSuperclass.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.0", ngImport: i0, type: DirectiveSuperclass, deps: null, target: i0.ɵɵFactoryTarget.Directive });
DirectiveSuperclass.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "14.0.0", version: "15.0.0", type: DirectiveSuperclass, selector: "[slDirectiveSuperclass]", usesInheritance: true, usesOnChanges: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.0", ngImport: i0, type: DirectiveSuperclass, decorators: [{
            type: Directive,
            args: [{ selector: '[slDirectiveSuperclass]' }]
        }] });

var _FormComponentSuperclass_changeDetectorRef;
/**
 * Use in the `providers` of a component that implements `ControlValueAccessor` to reduce some boilerplate.
 *
 * ```ts
 * @Component({ providers: [provideValueAccessor(MyFormControl)] }
 * class MyFormControl extends BaseFormControl {
 *   // ...
 * }
 * ```
 */
function provideValueAccessor(type) {
    return {
        provide: NG_VALUE_ACCESSOR,
        useExisting: type,
        multi: true,
    };
}
/**
 * Extend this when creating a form control to reduce some boilerplate.
 *
 * This example allows 2-way binding to a number via `[(ngModel)]`, `[formControl]`, or any other technique that leverages the `ControlValueAccessor` interface.
 * ```ts
 * @Component({
 *   template: `
 *     <button (click)="increment()" [disabled]="isDisabled">{{ counter }}</button>
 *   `,
 *   providers: [provideValueAccessor(CounterComponent)],
 * })
 * class CounterComponent extends FormComponentSuperclass<number> {
 *   counter = 0;
 *
 *   handleIncomingValue(value: number) {
 *     this.counter = value;
 *   }
 *
 *   increment() {
 *     this.emitOutgoingValue(++this.counter);
 *     this.onTouched();
 *   }
 * }
 * ```
 */
class FormComponentSuperclass extends DirectiveSuperclass {
    constructor() {
        super(...arguments);
        /** Call this to emit a new value when it changes. */
        this.emitOutgoingValue = noop;
        /** Call this to "commit" a change, traditionally done e.g. on blur. */
        this.onTouched = noop;
        /** You can bind to this in your template as needed. */
        this.isDisabled = false;
        _FormComponentSuperclass_changeDetectorRef.set(this, inject(ChangeDetectorRef));
    }
    /** Called as angular propagates value changes to this `ControlValueAccessor`. You normally do not need to use it. */
    writeValue(value) {
        this.handleIncomingValue(value);
        __classPrivateFieldGet(this, _FormComponentSuperclass_changeDetectorRef, "f").markForCheck();
    }
    /** Called as angular sets up the binding to this `ControlValueAccessor`. You normally do not need to use it. */
    registerOnChange(fn) {
        this.emitOutgoingValue = fn;
    }
    /** Called as angular sets up the binding to this `ControlValueAccessor`. You normally do not need to use it. */
    registerOnTouched(fn) {
        this.onTouched = fn;
    }
    /** Called as angular propagates disabled changes to this `ControlValueAccessor`. You normally do not need to use it. */
    setDisabledState(isDisabled) {
        this.isDisabled = isDisabled;
        __classPrivateFieldGet(this, _FormComponentSuperclass_changeDetectorRef, "f").markForCheck();
    }
}
_FormComponentSuperclass_changeDetectorRef = new WeakMap();

var _ControlSynchronizer_instances, _ControlSynchronizer_errorsFromPartner, _ControlSynchronizer_originalSetErrors, _ControlSynchronizer_finishingAsync, _ControlSynchronizer_acceptErrorsFrom, _ControlSynchronizer_setAddedErrors, _ControlSynchronizer_getNativeErrors, _ControlSynchronizer_updateCombinedErrors;
class ControlSynchronizer {
    constructor(control) {
        this.control = control;
        _ControlSynchronizer_instances.add(this);
        _ControlSynchronizer_errorsFromPartner.set(this, {});
        _ControlSynchronizer_originalSetErrors.set(this, this.control.setErrors.bind(this.control));
        _ControlSynchronizer_finishingAsync.set(this, false);
        wrapMethod(control, 'updateValueAndValidity', {
            around: (original, ...args) => {
                // when finishing async validators we instructed our partner to re-run its own validators. That could in turn try to re-trigger ours, dropping the errors we just set, putting us into a pending state, and re-triggering our own async validators that just completed.
                if (!__classPrivateFieldGet(this, _ControlSynchronizer_finishingAsync, "f")) {
                    original.apply(control, args);
                }
            },
        });
        wrapMethod(control, 'setErrors', {
            // ignore the second "opts" parameter that could set `emitEvent: false`. When async validation completes, we need to ensure a `statusChange` fires so we synchronize with `#partner`
            around: (_, newNativeErrors) => {
                __classPrivateFieldSet(this, _ControlSynchronizer_finishingAsync, true, "f");
                __classPrivateFieldGet(this, _ControlSynchronizer_instances, "m", _ControlSynchronizer_updateCombinedErrors).call(this, newNativeErrors);
                __classPrivateFieldSet(this, _ControlSynchronizer_finishingAsync, false, "f");
            },
        });
        control.addValidators(() => __classPrivateFieldGet(this, _ControlSynchronizer_errorsFromPartner, "f"));
    }
    static synchronize(outer, inner, outerToInnerTransform, innerToOuterTransform, subscriptionManager) {
        const outerSync = new ControlSynchronizer(outer);
        const innerSync = new ControlSynchronizer(inner);
        __classPrivateFieldGet(outerSync, _ControlSynchronizer_instances, "m", _ControlSynchronizer_acceptErrorsFrom).call(outerSync, innerSync, innerToOuterTransform, subscriptionManager);
        __classPrivateFieldGet(innerSync, _ControlSynchronizer_instances, "m", _ControlSynchronizer_acceptErrorsFrom).call(innerSync, outerSync, outerToInnerTransform, subscriptionManager);
    }
}
_ControlSynchronizer_errorsFromPartner = new WeakMap(), _ControlSynchronizer_originalSetErrors = new WeakMap(), _ControlSynchronizer_finishingAsync = new WeakMap(), _ControlSynchronizer_instances = new WeakSet(), _ControlSynchronizer_acceptErrorsFrom = function _ControlSynchronizer_acceptErrorsFrom(partner, transform, subscriptionManager) {
    subscriptionManager.subscribeTo(merge(partner.control.statusChanges, Promise.resolve()).pipe(map(() => __classPrivateFieldGet(partner, _ControlSynchronizer_instances, "m", _ControlSynchronizer_getNativeErrors).call(partner)), distinctUntilChanged(isEqual), transform), (transformedErrors) => {
        __classPrivateFieldGet(this, _ControlSynchronizer_instances, "m", _ControlSynchronizer_setAddedErrors).call(this, transformedErrors);
    });
}, _ControlSynchronizer_setAddedErrors = function _ControlSynchronizer_setAddedErrors(errorsToAdd) {
    __classPrivateFieldSet(this, _ControlSynchronizer_errorsFromPartner, errorsToAdd, "f");
    this.control.updateValueAndValidity({ emitEvent: false });
}, _ControlSynchronizer_getNativeErrors = function _ControlSynchronizer_getNativeErrors() {
    // TODO: remove the `!` after https://github.com/simontonsoftware/s-libs/issues/78
    return omit(this.control.errors, ...keys(__classPrivateFieldGet(this, _ControlSynchronizer_errorsFromPartner, "f")));
}, _ControlSynchronizer_updateCombinedErrors = function _ControlSynchronizer_updateCombinedErrors(nativeErrors) {
    const combinedErrors = { ...nativeErrors, ...__classPrivateFieldGet(this, _ControlSynchronizer_errorsFromPartner, "f") };
    __classPrivateFieldGet(this, _ControlSynchronizer_originalSetErrors, "f").call(this, size(combinedErrors) ? combinedErrors : null);
};

var _WrappedControlSuperclass_instances, _WrappedControlSuperclass_incomingValues$, _WrappedControlSuperclass_injector, _WrappedControlSuperclass_errorHandler, _WrappedControlSuperclass_bindValidation, _WrappedControlSuperclass_handleError, _WrappedControlSuperclass_selfInject;
/**
 * Extend this when creating a form component that simply wraps existing ones, to reduce a lot of boilerplate.
 *
 * To wrap a single form control use the subclass {@linkcode WrappedFormControlSuperclass}:
 * ```ts
 * @Component({
 *   template: `<input [formControl]="control">`,
 *   providers: [provideValueAccessor(StringComponent)],
 * })
 * class StringComponent extends WrappedFormControlSuperclass<string> {}
 * ```
 *
 * Example of wrapping multiple inner components:
 * ```ts
 * class FullName {
 *   firstName = '';
 *   lastName = '';
 * }
 *
 * @Component({
 *   selector: 'app-full-name',
 *   template: `
 *     <div [formGroup]="control">
 *       <input id="first" formControlName="firstName" />
 *       <input id="last" formControlName="lastName" />
 *     </div>
 *   `,
 *   providers: [provideValueAccessor(FullNameComponent)],
 * })
 * class FullNameComponent extends WrappedControlSuperclass<FullName> {
 *   control = new FormGroup({
 *     firstName: new FormControl(),
 *     lastName: new FormControl(),
 *   });
 *
 *   protected outerToInnerValue(outer: FullName | null): FullName {
 *     // `outer` can come in as `null` during initialization when the user binds with `ngModel`
 *     return outer || new FullName();
 *   }
 * }
 * ```
 *
 * Example when you need to modify the wrapped control's value:
 * ```ts
 * @Component({
 *   template: `<input type="datetime-local" [control]="formControl">`,
 *   providers: [provideValueAccessor(DateComponent)],
 * })
 * class DateComponent extends WrappedFormControlSuperclass<Date, string> {
 *   protected innerToOuterValue(inner: string): Date {
 *     return new Date(inner + "Z");
 *   }
 *
 *   protected outerToInnerValue(outer: Date): string {
 *     if (outer === null) {
 *       return ""; // happens during initialization
 *     }
 *     return outer.toISOString().substr(0, 16);
 *   }
 * }
 * ```
 *
 * If you bind to your component using an {@linkcode NgControl} (e.g. when using `ngModel`), validation errors will be synchronized between it and the control inside your component. You can override various methods below to control or disable that process. Note that validation, `statuschanges`, and `valuechanges` may all happen more often as a result of this synchronization.
 */
class WrappedControlSuperclass extends FormComponentSuperclass {
    constructor() {
        super();
        _WrappedControlSuperclass_instances.add(this);
        _WrappedControlSuperclass_incomingValues$.set(this, new Subject());
        _WrappedControlSuperclass_injector.set(this, inject(Injector));
        _WrappedControlSuperclass_errorHandler.set(this, inject(ErrorHandler));
        this.subscribeTo(this.setUpOuterToInnerValue$(__classPrivateFieldGet(this, _WrappedControlSuperclass_incomingValues$, "f")), (inner) => {
            this.control.setValue(inner, { emitEvent: false });
        });
    }
    ngOnInit() {
        __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_bindValidation).call(this);
        this.subscribeTo(this.setUpInnerToOuterValue$(this.control.valueChanges), (outer) => {
            this.emitOutgoingValue(outer);
        });
        wrapMethod(this.control, 'markAsTouched', {
            after: () => {
                this.onTouched();
            },
        });
    }
    /** Called as angular propagates values changes to this `ControlValueAccessor`. You normally do not need to use it. */
    handleIncomingValue(outer) {
        __classPrivateFieldGet(this, _WrappedControlSuperclass_incomingValues$, "f").next(outer);
    }
    /** Called as angular propagates disabled changes to this `ControlValueAccessor`. You normally do not need to use it. */
    setDisabledState(isDisabled) {
        if (isDisabled) {
            this.control.disable({ emitEvent: false });
        }
        else {
            this.control.enable({ emitEvent: false });
        }
        super.setDisabledState(this.isDisabled);
    }
    /**
     * Override this to for full control over the stream of values passed in to your subclass.
     *
     * In this example, incoming values are debounced before being passed through to the inner form control
     * ```ts
     * setUpOuterToInnerValue$(outer$: Observable<OuterType>): Observable<InnerType> {
     *   return outer$.pipe(
     *     debounce(300),
     *     map((outer) => doExpensiveTransformToInnerValue(outer)),
     *   );
     * }
     * ```
     *
     * For a simple transformation, see {@linkcode #outerToInnerValue} instead.
     */
    setUpOuterToInnerValue$(outer$) {
        return outer$.pipe(map((outer) => this.outerToInnerValue(outer)), __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_handleError).call(this));
    }
    /**
     * Override this to modify a value coming from the outside to the format needed within this component.
     *
     * For more complex needs, see {@linkcode #setUpOuterToInnerValue$} instead.
     */
    outerToInnerValue(outer) {
        return outer;
    }
    /**
     * Override this to for full control over the stream of values emitted from your subclass.
     *
     * In this example, illegal values are not emitted
     * ```ts
     * setUpInnerToOuterValue$(inner$: Observable<InnerType>): Observable<OuterType> {
     *   return inner$.pipe(
     *     filter((inner) => isLegalValue(outer)),
     *   );
     * }
     * ```
     *
     * For a simple transformation, see {@linkcode #innerToOuterValue} instead.
     */
    setUpInnerToOuterValue$(inner$) {
        return inner$.pipe(map((inner) => this.innerToOuterValue(inner)), __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_handleError).call(this));
    }
    /**
     * Override this to modify a value coming from within this component to the format expected on the outside.
     *
     * For more complex needs, see {@linkcode #setUpInnerToOuterValue$} instead.
     */
    innerToOuterValue(inner) {
        return inner;
    }
    /**
     * Override this to for full control over the stream of validation errors synchronized in to your subclass.
     *
     * For a simple transformation, see {@linkcode #outerToInnerErrors} instead.
     */
    setUpOuterToInnerErrors$(outer$) {
        return outer$.pipe(map((inner) => this.outerToInnerErrors(inner)), __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_handleError).call(this));
    }
    /**
     * Override this to modify validation errors that synchronize in to this component.
     *
     * In this example we assume the `required` validation is handled by the user and should not affect internal validation
     * ```ts
     * protected override outerToInnerErrors(
     *   errors: ValidationErrors,
     * ): ValidationErrors {
     *   return omit(errors, 'required');
     * }
     * ```
     *
     * For more complex needs, see {@linkcode #setUpOuterToInnerErrors$} instead.
     */
    outerToInnerErrors(errors) {
        return errors;
    }
    /**
     * Override this to for full control over the stream of validation errors synchronized out from your subclass.
     *
     * In this example, synchronization is turned off:
     * ```ts
     * protected override setUpInnerToOuterErrors$(): Observable<ValidationErrors> {
     *   return EMPTY;
     * }
     * ```
     *
     * For a simple transformation, see {@linkcode #innerToOuterErrors} instead.
     */
    setUpInnerToOuterErrors$(inner$) {
        return inner$.pipe(map((inner) => this.innerToOuterErrors(inner)), __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_handleError).call(this));
    }
    /**
     * Override this to modify validation errors that synchronize out from this component.
     *
     * For more complex needs, see {@linkcode #setUpInnerToOuterErrors$} instead.
     */
    innerToOuterErrors(errors) {
        return errors;
    }
}
_WrappedControlSuperclass_incomingValues$ = new WeakMap(), _WrappedControlSuperclass_injector = new WeakMap(), _WrappedControlSuperclass_errorHandler = new WeakMap(), _WrappedControlSuperclass_instances = new WeakSet(), _WrappedControlSuperclass_bindValidation = async function _WrappedControlSuperclass_bindValidation() {
    // Hack-fixing a production bug: https://github.com/simontonsoftware/s-libs/issues/82. `ngModel` and `formControl` both set their `.control` before this component is initialized. However, `formControlName` does not. This is a timing hack to accommodate by delaying on the microtask queue.
    await Promise.resolve();
    const outerControl = __classPrivateFieldGet(this, _WrappedControlSuperclass_instances, "m", _WrappedControlSuperclass_selfInject).call(this, NgControl)?.control;
    if (outerControl) {
        ControlSynchronizer.synchronize(outerControl, this.control, this.setUpOuterToInnerErrors$.bind(this), this.setUpInnerToOuterErrors$.bind(this), this);
    }
}, _WrappedControlSuperclass_handleError = function _WrappedControlSuperclass_handleError() {
    return flow(tap({ error: bindKey(__classPrivateFieldGet(this, _WrappedControlSuperclass_errorHandler, "f"), 'handleError') }), retry());
}, _WrappedControlSuperclass_selfInject = function _WrappedControlSuperclass_selfInject(token) {
    return __classPrivateFieldGet(this, _WrappedControlSuperclass_injector, "f").get(token, undefined, { self: true, optional: true });
};
WrappedControlSuperclass.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "15.0.0", ngImport: i0, type: WrappedControlSuperclass, deps: [], target: i0.ɵɵFactoryTarget.Directive });
WrappedControlSuperclass.ɵdir = i0.ɵɵngDeclareDirective({ minVersion: "14.0.0", version: "15.0.0", type: WrappedControlSuperclass, usesInheritance: true, ngImport: i0 });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "15.0.0", ngImport: i0, type: WrappedControlSuperclass, decorators: [{
            type: Directive
        }], ctorParameters: function () { return []; } });

/**
 * This is a convenience class to make it easier to extend {@linkcode WrappedControlSuperclass} when you specifically want a `FormControl` (as opposed to a `FormArray` or `FormGroup`).
 */
class WrappedFormControlSuperclass extends WrappedControlSuperclass {
    constructor() {
        super(...arguments);
        this.control = new FormControl();
    }
}

/*
 * Public API Surface of ng-core
 */

/**
 * Generated bundle index. Do not edit.
 */

export { DirectiveSuperclass, FormComponentSuperclass, InjectableSuperclass, WrappedControlSuperclass, WrappedFormControlSuperclass, mixInInjectableSuperclass, provideValueAccessor };
//# sourceMappingURL=ng-core.mjs.map
