import { OnChanges, SimpleChanges } from '@angular/core';
import { Observable, Subject } from 'rxjs';
import { InjectableSuperclass } from './injectable-superclass';
import * as i0 from "@angular/core";
/**
 * Extend this when creating a directive (including a component, which is a kind of directive) to gain access to the helpers demonstrated below.
 *
 * ```ts
 * @Component({
 *   selector: "s-color-text",
 *   template: `
 *     <span [style.background]="color">{{ color }}</span>
 *   `,
 *   // note that `bindToInstance()` works even with OnPush change detection
 *   changeDetection: ChangeDetectionStrategy.OnPush,
 * })
 *  class ColorTextComponent extends DirectiveSuperclass {
 *   @Input() prefix?: string;
 *   @Input() prefix2?: string;
 *   color!: string;
 *
 *   constructor(@Inject("color$") color$: Observable<string>) {
 *     super();
 *
 *     // combine everything to calculate `color` and keep it up to date
 *     this.bindToInstance(
 *       "color",
 *       combineLatest(
 *         this.getInput$("prefix"),
 *         this.getInput$("prefix2"),
 *         color$,
 *       ).pipe(map((parts) => parts.filter((p) => p).join(""))),
 *     );
 *   }
 * }
 * ```
 */
export declare abstract class DirectiveSuperclass extends InjectableSuperclass implements OnChanges {
    #private;
    /**
     *  Emits the set of `@Input()` property names that change during each call to `ngOnChanges()`.
     */
    inputChanges$: Subject<Set<keyof this>>;
    ngOnChanges(changes: SimpleChanges): void;
    /**
     * @return an observable of the values for one of this directive's `@Input()` properties
     */
    getInput$<K extends keyof this>(key: K): Observable<this[K]>;
    /**
     * Binds an observable to one of this directive's instance variables. When the observable emits the instance variable will be updated, and change detection will be triggered to propagate any changes. Use this an an alternative to repeating `| async` multiple times in your template.
     */
    bindToInstance<K extends keyof this>(key: K, value$: Observable<this[K]>): void;
    static ɵfac: i0.ɵɵFactoryDeclaration<DirectiveSuperclass, never>;
    static ɵdir: i0.ɵɵDirectiveDeclaration<DirectiveSuperclass, "[slDirectiveSuperclass]", never, {}, {}, never, never, false, never>;
}
